#!/bin/sh -
#	$NetBSD: builtins.def,v 1.15 2000/04/09 23:27:03 christos Exp $
#
# Copyright (c) 1991, 1993
#	The Regents of the University of California.  All rights reserved.
#
# This code is derived from software contributed to Berkeley by
# Kenneth Almquist.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
# 3. All advertising materials mentioning features or use of this software
#    must display the following acknowledgement:
#	This product includes software developed by the University of
#	California, Berkeley and its contributors.
# 4. Neither the name of the University nor the names of its contributors
#    may be used to endorse or promote products derived from this software
#    without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
# OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
# OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
# SUCH DAMAGE.
#
#	@(#)builtins.def	8.4 (Berkeley) 5/4/95

#
# This file lists all the builtin commands.  The first column is the name
# of a C routine.  The -j flag, if present, specifies that this command
# is to be excluded from systems without job control, and the -h flag,
# if present specifies that this command is to be excluded from systems
# based on the SMALL compile-time symbol.  The rest of the line
# specifies the command name or names used to run the command.  The entry
# for bltincmd, which is run when the user does not specify a command, must
# come first.
#
# NOTE: bltincmd must come first!

bltincmd	command
#alloccmd	alloc
bgcmd -j	bg
breakcmd	break continue
#catfcmd	catf
cdcmd		cd chdir
dotcmd		.
echocmd		echo
evalcmd		eval
execcmd		exec
exitcmd		exit
expcmd		exp let
exportcmd	export readonly
#exprcmd		expr
falsecmd	false
histcmd -h	fc
fgcmd -j	fg
getoptscmd	getopts
hashcmd		hash
jobidcmd	jobid
jobscmd		jobs
#linecmd		line
localcmd	local
#nlechocmd	nlecho
#printfcmd	printf
pwdcmd		pwd
readcmd		read
returncmd	return
setcmd		set
setvarcmd	setvar
shiftcmd	shift
trapcmd		trap
truecmd		: true
typecmd		type
umaskcmd	umask
unaliascmd	unalias
unsetcmd	unset
waitcmd		wait
#foocmd		foo
aliascmd	alias
ulimitcmd	ulimit
testcmd		test [
